"""Schema compilation: Finite State Machine implementation
for content models
"""

__version__="$Revision: 1.12 $"
# $Id: FSM.py,v 1.12 2004/05/20 09:50:20 ht Exp $

# FSM for schema
import copy
import types
import sys

from Group import All, Sequence, Choice
from Element import Element
from Wildcard import Wildcard
from QName import QName
from XSV.infoset.XMLInfoset import Element as ISElement
from XSV.infoset.XMLInfoset import Characters
#from XSV.validate.verror import verror

import Particle

class FSM:
  def __init__(self):
    self.nodes = []
    self.eltTab = {}
    self.startNode = None
    self.counters = []
    self.npTab = {}
    self.nextID = 1

  def initCounters(self):
#    print 'ic'
    for c in self.counters:
      c.init()

  def addCounter(self,cnt):
    self.counters.append(cnt)
    
  def assignIDs(self,reverse=1):
    if not self.nodes:
      return
    if reverse:
      self.nodes.reverse()
    for n in self.nodes:
      if n.id:
        continue
      #    print "assigning id %d to %s" % (self.fsm.nextID,self)
      n.id = self.nextID
      self.nextID = self.nextID + 1

  def printme(self,file):
    self.assignIDs()
    self.nodes.sort(FSMNode.compareIDs)
    for n in self.nodes:
#      try:
      file.write("%2d" % n.id)
#      except:
 #       print ('fw',n,n.id)
  #      xyzzy()
      if n.label is not None:
        file.write(" ");
        if not n.label[0].node.__dict__.has_key('id'):
          n.label[0].node.fsm.assignIDs()
        sep="("
        for m in n.label:
          file.write(sep);
          file.write("%s"%unicode(m));
          sep=", "
        file.write(")")
      file.write(":\n")
#      if n.label:
#  	sys.stdout.write("[")
#  	for nn in n.label:
#  	  sys.stdout.write("%d " % nn.id)
#  	sys.stdout.write("]")
      for e in n.edges:
        e.printme(file)
        file.write("\n");
        
  def asXML(self):
    res=ISElement(None,None,"fsm")
    self.assignIDs(0)
    self.nodes.sort(FSMNode.compareIDs)
    for n in self.nodes:
      ne=res.newDaughter("node")
      ne.newAttr('id',("%s" % n.id))
      nen=0
      if n.isEndNode():
	ne.newAttr('final','true')
      else:  
        n.edges.sort(FSMEdge.compare)
      if len(n.counters)>0:
        ne.newAttr('cnt',str(map(lambda c:unicode(c),n.counters)))
      for e in n.edges:
        if e.__class__ is ExitEdge:
          fd=e.descC()
          if fd!="":
            ne.newAttr('actions',fd)
          continue
        ee=ne.newDaughter("edge")
        lab=unicode(e)
        ed=e.descC()
        if ed!="":
          ee.newAttr('actions',ed)
        ee.newAttr('label',lab)
        ee.newAttr('dest',("%s" % e.dest.id))
    return res

  # cf Aho & Ullman p93
  # but note we never have two identical labels, so we only
  # have to worry about lambda arcs
  def determinise(self):
    D = FSM()
    D.counters = self.counters
    D.startNode = FSMNode(D)
    D.startNode.label = _eclosure(self.startNode)
    D.unmarkedNodes = [D.startNode]
    while D.unmarkedNodes:
      x = D.unmarkedNodes.pop()
      destnodes = {}
      for n in x.label:
        checkers = n.checkers
        n = n.node
	for e in n.edges:
          if e.__class__ is not EmptyEdge:
            if e.__class__ is ExitEdge:
              ee=e.clone(x)
            else:
              y = _eclosure(e.dest)
              for n in D.nodes:
                if y == n.label:
                  break
              else:
                n = FSMNode(D)
                D.unmarkedNodes.append(n)
                n.label = y
              ee=e.clone(x, n)
            ee.checkers=ee.checkers+checkers
    return D

  def subsumed(self, fsm):
    # check whether self is subsumed by fsm, by parsing all paths thru self
    unprocStates = [(fsm.startNode,self.startNode)]
    procStates = []
    while unprocStates:
      p=unprocStates[0]
      b=p[0]
      d=p[1]
      for de in d.edges:
        win=0
        for be in b.edges:
          sr=de.subsumed(be)
          if type(sr)==types.StringType:
            return ('notSubType',de,d,b,sr)
          if sr:
            if sr==2:
              # two element edges
              # what about value/identity constraints?
              myDec=de.decl
              otherDec=be.decl
              if (myDec is not None and
                  otherDec is not None and
                  (type(myDec.scope) is not types.StringType or
                   type(otherDec.scope) is not types.StringType) and
                  not myDec.typeDefinition.isSubtype(otherDec.typeDefinition,['extension']+otherDec.typeDefinition.final)):
                return ('notSubType',de,d,b,be)
            win=1
            # check end status
            if sr!=3:
              ns=(be.dest,de.dest)
              if ns not in procStates:
                if ns not in unprocStates:
                  unprocStates.append(ns)
            break
        if not win:
          return ('noparse',de,d,b)
      procStates.append(p)
      unprocStates=unprocStates[1:]
    # success

  def subsumptionError(self,badGuy,other,checkResult,variety,rName,bName):
    if len(checkResult)==3:
      (rType,nodeR,nodeB) = checkResult
    elif len(checkResult)==4:
      (rType,edgeR,nodeR,nodeB) = checkResult
    else:
      (rType,edgeR,nodeR,nodeB,extra) = checkResult
      
    pics=[Characters(None,"\nBase model:\n"),
          other.asXML(),
          Characters(None,"\nThis model:\n"),
          self.asXML()]
    if rType=='noparse':
      lab=unicode(edgeR)
      badGuy.error("Content model of %s not actually a restriction of base %s %s: a %s element at node %s of this model can't be matched by any edge leaving node %s of the base"%(rName,variety,bName,lab,nodeR.id,nodeB.id),0,pics)
    elif rType=='notSubType':
      lab=unicode(edgeR)
      dtd=edgeR.decl.typeDefinition.name
      if type(extra)==types.StringType:
        badGuy.error("Content model of %s not actually a restriction of base %s %s: the %s element at node %s of this model has a type (%s) not derived from the type (%s) of the global element accepted by the matching wildcard edge leaving node %s of the base"%(rName,variety,bName,lab,nodeR.id,dtd,extra,nodeB.id),0,pics)
      else:
        btd=extra.decl.typeDefinition.name
        badGuy.error("Content model of %s not actually a restriction of base %s %s: the %s element at node %s of this model has a type (%s) not derived from the type (%s) of the matching %s edge leaving node %s of the base"%(rName,variety,bName,lab,nodeR.id,dtd,btd,lab,nodeB.id),0,pics)
    elif rType=='notEnd':
      badGuy.error("Content model of %s not actually a restriction of base %s %s: no way forward from this models' end state %s from the base's node %s which gets to an end state"%(rName,variety,bName,nodeR.id,nodeB.id),0,pics)
    else:
      shouldnt('notSubsumed')

  def newNodePlus(self,node,edgelist):
    if len(edgelist)>0:
      # should this be sorted?
      kk = tuple(edgelist+[node])
    else:
      kk=node
    try:
      return self.npTab[kk]
    except:
      nn=NodePlus(node,edgelist)
      self.npTab[kk]=nn
      return nn

class NodePlus:
  def __init__(self,node,edgelist):
    self.node = node
    self.checkers = []
    for edge in edgelist:
      self.checkers.extend(edge.checkers)

  def __str__(self):
    sim="%d"%self.node.id
    sep="("
    if len(self.checkers)!=0:
      for c in self.checkers:
        sim=sim+sep
        sim=sim+c.toString(0)
        sep=", "
    if sep=="(":
      return sim
    else:
      return sim+")"

class Counter:
  idBase='a'
  def __init__(self,particle,fsm,min,max):
    fsm.addCounter(self)
    self.min=min
    self.max=max
    self.particle=particle
    self.id=self.idBase
    Counter.idBase=chr(ord(Counter.idBase)+1)

  def __str__(self):
    return self.toString(0)

  def init(self):
    self.count = 0

  def incr(self):
    self.count = self.count+1
    return self.max is None or self.count<=self.max
    
  def checkMin(self):
    return self.count>=self.min

  def maxedOut(self):
    return self.max==self.count
  
  def toString(self,counter):
    if counter:
      return "%s++"%self.id
    else:
      return "%s!"%self.id

class FSMEdge:
  def __init__(self, source):
    self.checkers=[]
    self.source = source
    source.addEdge(self)

  def compare(self,other):
    if self.rank<other.rank:
      return -1
    elif self.rank>other.rank:
      return 1
    else:
      return self.compare(other)

  def addChecker(self,cnt):
    self.checkers.append(cnt)

  def descC(self):
    first=1
    res=""
    for c in self.checkers:
      if not first:
        res=res+";"+c.toString(0)
      else:
        res=res+c.toString(0)
      first=0
    return res

  def printC(self,file):
    for c in self.checkers:
      file.write(";")
      file.write(c.toString(0))

class RealEdge(FSMEdge):
  # persists in determinised automata
  def __init__(self,source):
    FSMEdge.__init__(self,source)

  def checkMins(self):
    for c in self.checkers:
      if not c.checkMin():
        c.errmsg="Min not reached: %s"%c.min
        return False
    return True

  def initMins(self):
    for c in self.checkers:
      c.init()

  def maxedOut(self):
    return self.source.maxedOut()

  def doCounters(self):
    if self.checkMins():
      self.initMins()
      return True
    else:
      return False

  def subsumed(self,other):
    # is this edge subsumed by other
    if (other.__class__ is ExitEdge or
        other.__class__ is IncrEdge):
      # we're not, or E/I E.subsumed would have caught this
      return 0
    elif other.__class__ is ElementEdge:
      return self.subsumedByE(other)
    elif other.__class__ is WildEdge:
      return self.subsumedByW(other)
    else:
      return self.subsumedByM(other)

  def overlap(self,other):
    # does this edge overlap with other
    if (other.__class__ is ExitEdge or
        other.__class__ is IncrEdge):
      # we're not, or E/I E.overlap would have caught this
      return
    elif other.__class__ is ElementEdge:
      return self.overlapWithE(other)
    elif other.__class__ is WildEdge:
      return self.overlapWithW(other)
    else:
      return self.overlapWithM(other)

  def clone(self,other):
    other.checkers=self.checkers
  
  def match(self,c):
    return
  
  def matchEnd(self):
    return
  
class EmptyEdge(FSMEdge):
  rank = 10
  def __init__(self,source,dest):
    FSMEdge.__init__(self,source)
    self.dest=dest

  def clone(self,source,dest):
    e=EmptyEdge(source,dest)
    RealEdge.clone(self,e)
    return e

  def printme(self,file):
    file.write("    L->%d" % self.dest.id)
    self.printC(file);

  def compare(self,other):
    return 0

class ExitEdge(RealEdge):
  rank = 6
  dest = None

  def __init__(self,source):
    RealEdge.__init__(self,source)

  def __str__(self):
    return "$"

  def clone(self,source):
    e=ExitEdge(source)
    RealEdge.clone(self,e)
    return e

  def printme(self,file):
    file.write("    $->")
    self.printC(file);

  def subsumed(self,other):
    if other.__class__ is ExitEdge:
      return 3
    else:
      return 0

  def compare(self,other):
    return 0

  def matchEnd(self):
    return ((not self.maxedOut()) and
            self.doCounters())

  def overlap(self,other):
    return

class TestEdge(RealEdge):
  def __init__(self,source,dest,decl):
    RealEdge.__init__(self,source)
    self.dest=dest
    self.decl=decl

  def printme(self,file):
    file.write("    %s->%d"%(unicode(self),self.dest.id))
    self.printC(file);

class ElementEdge(TestEdge):
  rank = 2
  def __init__(self,qname,decl,source,dest):
    TestEdge.__init__(self,source,dest,decl)
    self.qname=qname
    self.uri=self.qname.uri or None
    self.local=self.qname.local

  def __str__(self):
    if self.decl is None:
      # no decl yet
      return str(self.qname)+"*"
    else:
      return str(self.qname)

  def clone(self,source,dest):
    e=ElementEdge(self.qname,self.decl,source,dest)
    RealEdge.clone(self,e)
    return e

  def compare(self,other):
    if self.qname < other.qname:
      return -1
    elif self.qname > other.qname:
      return +1
    else:
      return 0

  def match(self,c):
    if ((not self.maxedOut()) and
        c.namespaceName==self.uri and
        c.localName==self.local):
      if self.doCounters():
        return self.dest
    
  def subsumedByE(self,other):
    if self.qname == other.qname:
      return 2
    else:
      return 0

  def subsumedByW(self,other):
    bw=other.decl                 # element (self) vs. wildcard (bw)
    if bw.allows(self.uri or None):
      # Wildcard vis a vis explicit decl -- by cases wrt processContents
      if bw.processContents=="skip":
        return 1
      else:
        # I don't _think_ it makes sense to tryHardForDecl here -- there
        # might not _be_ any instance
        if (bw.schema.vElementTable.has_key(self.qname) and
            bw.schema.vElementTable[self.qname] is not None):
          gd=bw.schema.vElementTable[self.qname]
          if (self.decl is not None and
              self.decl.typeDefinition.isSubtype(gd.typeDefinition,['extension']+gd.typeDefinition.final)):
            return 1
          else:
            return gd.typeDefinition.name
        else:
          # lose if strict, as base wouldn't allow elt (no global defn)
          return bw.processContents=="lax"
    else:
      return 0

  def overlapWithW(self,other):
    return other.overlapWithE(self)

  def overlapWithE(self,other):
    if self.qname==other.qname:
      return "%s/%s" % (self.qname,other.qname)

  def overlapWithM(self,other):
    pass

class WildEdge(TestEdge):
  rank = 3
  def __init__(self,decl,source,dest):
    TestEdge.__init__(self,source,dest,decl)

  def __str__(self):
    return str(self.decl)

  def clone(self,source,dest):
    e=WildEdge(self.decl,source,dest)
    RealEdge.clone(self,e)
    return e

  def compare(self,other):
    if self.decl.allowed < other.decl.allowed:
      return -1
    elif self.decl.allowed > other.decl.allowed:
      return +1
    else:
      return 0

  def match(self,c):
    if ((not self.maxedOut()) and
        self.decl.allows(c.namespaceName or None)):
      if self.doCounters():
        return self.dest

  def subsumedByE(self,other):
    return 0

  def subsumedByW(self,other):
    return self.decl.subsumed(other.decl)

  def overlapWithW(self,other):
    if not self.decl.intersect(other.decl).isEmpty():
      return "%s/%s" % (self.decl,other.decl)

  def overlapWithE(self,other):
    if self.decl.allows(other.uri):
      return "%s/%s" % (self.decl,other.qname)

  def overlapWithM(self,other):
    pass

class MaxEdge(TestEdge):
  rank = 4
  def __init__(self,cnt,source,dest):
    TestEdge.__init__(self,source,dest,None)
    self.cnt=cnt

  def __str__(self):
    return "M(%s)"%self.cnt.id

  def clone(self,source,dest):
    e=MaxEdge(self.cnt,source,dest)
    RealEdge.clone(self,e)
    return e

  def compare(self,other):
    if self.cnt.id < other.cnt.id:
      return -1
    elif self.cnt.id > other.cnt.id:
      return +1
    else:
      return 0

  def matchEnd(self):
    return self.match(None)

  def match(self,c):
    if self.maxedOut():
      if self.doCounters():
        self.cnt.init()
        return self.dest

  def subsumedByW(self,other):
    return 0

  def subsumedByE(self,other):
    return 0

  def subsumedByM(self,other):
    return other.cnt is self.cnt

  def overlapWithW(self,other):
    pass

  def overlapWithE(self,other):
    pass

  def overlapWithM(self,other):
    if self.cnt is other.cnt:
      return "%s/%s"%(unicode(self),unicode(other))

class IncrEdge(TestEdge):
  rank = 5
  def __init__(self,cnt,source,dest):
    TestEdge.__init__(self,source,dest,None)
    self.cnt=cnt

  def __str__(self):
    return "++(%s)"%self.cnt.id

  def clone(self,source,dest):
    e=IncrEdge(self.cnt,source,dest)
    RealEdge.clone(self,e)
    dest.addCounter(self.cnt)
    return e

  def compare(self,other):
    if self.cnt.id < other.cnt.id:
      return -1
    elif self.cnt.id > other.cnt.id:
      return +1
    else:
      return 0

  def matchEnd(self):
    return self.match(None)

  def match(self,c):
    if not self.maxedOut():
      if self.cnt.incr():
        if self.doCounters():
          return self.dest
      else:
        self.cnt.errmsg="Max exceeded: %s"%self.cnt.max

  def subsumed(self,other):
    if other.__class__ is IncrEdge:
      return other.cnt is self.cnt
    else:
      return 0

  def overlap(self,other):
    if other.__class__ is IncrEdge:
      if self.cnt is other.cnt:
        return "%s/%s"%(unicode(self),unicode(other))

class FSMNode:
  id=None
  def __init__(self, fsm):
    fsm.nodes.append(self)
    self.fsm = fsm
    self.edges = []
    self.label = None
    self.mark = None
    self.counters = []

  def isEndNode(self):
    for e in self.edges:
      if e.__class__ is ExitEdge:
        return True
    return False

  def makeEndNode(self):
    ExitEdge(self)

  def compareIDs(self,other):
    if self.id < other.id:
      return -1
    elif self.id > other.id:
      return +1
    else:
      return 0

  def addEdge(self,edge):
    # This means Incr's come _last_, because
    # validateElementModel iterates backwards
    # That means we defer to a real edge,
    # (Only happens at beginning of loop?)
    # which is what makes us greedy
    if edge.__class__ is IncrEdge:
      self.edges.insert(0,edge)
    else:
      self.edges.append(edge)

  def addCounter(self,cnt):
    if cnt not in self.counters:
      self.counters.append(cnt)

  def maxedOut(self):
    for c in self.counters:
      if c.maxedOut():
        return True
    return False

# cf Aho & Ullman p92

def _eclosure(n):
  # we use the mark to accumulate all the counters involved in
  # reaching a particular state
  #print ('ec',n.id)
  fsm=n.fsm
  n.mark=[]
  STACK = [n]
  ECLOSURE = [fsm.newNodePlus(n,n.mark)]
  while STACK:
    s = STACK.pop()
    for e in s.edges:
      if e.__class__ is EmptyEdge:
	t = e.dest
	if t.mark is None:
          #          print ('ecy',t.id)
          if len(e.checkers)>0:
            nmark=s.mark+[e]
          else:
            nmark=s.mark
          t.mark=nmark
	  ECLOSURE.append(fsm.newNodePlus(t,nmark))
	  STACK.append(t)
#        else:
 #         print('ecn',t.id)
  ECLOSURE.sort() # so == will work
  for m in ECLOSURE:
    m.node.mark=None
  # print ('ec',n.id,map(lambda e:e.id,ECLOSURE))
  return ECLOSURE

# Check a FSM is deterministic

def _checkFSM(fsm):
  for n in fsm.nodes:
    for e in range(len(n.edges)):
      l = n.edges[e]
      for f in range(0,e):
        m = n.edges[f]
        r = l.overlap(m)
        if r is not None:
          return r

def _translateA(self, next, loopOverhead):
  # doesn't enforce cardinality yet
  n = FSMNode(next.fsm)
  for particle in self.particles:
    m = particle.translate(n, loopOverhead)
    EmptyEdge(n, m)
  EmptyEdge(n, next)
  return n

def _translateE(self, next, loopOverhead):
  # what do I do if there are no edges, i.e. abstract element with
  # no descendants?
  fsm = next.fsm
  eltTab = fsm.eltTab
  n = FSMNode(fsm)
  if self.abstract!='true':
    ElementEdge(self.qname,self, n, next)
  if eltTab.has_key(self.qname):
    if self.typeDefinition  is not  eltTab[self.qname].typeDefinition:
      self.error("illegal redeclaration of %s" % self.qname)
  else:
    eltTab[self.qname]=self
  if (type(self.scope) is types.StringType and
      'substitution' not in self.prohibitedSubstitutions):
    # used to be self.schema.vElementTable[qname].equivClass -- why???
    for e in self.equivClass:
      if e.abstract!='true':
        ElementEdge(e.qname, e, n, next)
      if eltTab.has_key(e.qname):
        if e.typeDefinition  is not  eltTab[e.qname].typeDefinition:
          self.error("illegal redeclaration of %s" % e.qname)
      else:
        eltTab[e.qname]=e
  return n

def _translateW(self, next, loopOverhead):
  n = FSMNode(next.fsm)
  WildEdge(self, n, next)
  return n

def _translateC(self, next, loopOverhead):
  n = FSMNode(next.fsm)
  for particle in self.particles:
    m = particle.translate(next,loopOverhead)
    EmptyEdge(n, m)
  return n

def _translateS(self, next,loopOverhead):
  n = next
  rmodel = copy.copy(self.particles)
  rmodel.reverse()
  for particle in rmodel:
    n = particle.translate(n,loopOverhead)
  return n

def _translateP(self, next, loopOverhead):
  fsm = next.fsm
  if self.term is None:
    # undefined ref?
    if self.termType=='element' and self.termName:
      # we can check the content model, recursive type check may fail
      qq=self.termName
    else:
      qq=Particle.UndefQName
  min=self.occurs[0]
  max=self.occurs[1]
  if (loopOverhead and not
      (min==0 or min==max)):
    self.error("violation of constraints on exponents",
               1)
  loop=min>1 or max is None or max > 1
  if loop:
    # need a loop, overkill if no counter needed
    cnt = None
    postIncr = FSMNode(fsm)
    subLast = FSMNode(fsm)
    if self.term is not None:
      subFirst = self.term.translate(subLast,True)
    else:
      subFirst = FSMNode(fsm)
      ElementEdge(qq, None, subFirst, subLast)
    first = FSMNode(fsm)
    EmptyEdge(first,subFirst)
    outEdge = EmptyEdge(first,next)
    EmptyEdge(postIncr,first)
    if min>1 or max is not None:
      # do counter stuff
      cnt = Counter(self,fsm,min,max)
      IncrEdge(cnt,subLast,postIncr)
      MaxEdge(cnt,postIncr,next)
      outEdge.addChecker(cnt)
    else:
      EmptyEdge(subLast,postIncr)
      EmptyEdge(postIncr,next)
  else:
    # either (0,1) or (1,1)
    if self.term is not None:
      first = self.term.translate(next,loopOverhead)
    else:
      first = FSMNode(fsm)
      ElementEdge(qq, None, first, next)
    if min==0:
      EmptyEdge(first, next)
  return first

def UniqueFSM(particle):
  ndfsm = FSM()
  end = FSMNode(ndfsm)
  ExitEdge(end)
  start = particle.translate(end,0)
  ndfsm.startNode = start
  #ndfsm.printme(sys.stdout)
  fsm = ndfsm.determinise()
  #fsm.assignIDs(0)
  #fsm.printme(sys.stdout)
  return (fsm,_checkFSM(fsm))

def init():
  Particle.Particle.translate=_translateP
  Sequence.translate=_translateS
  Choice.translate=_translateC
  Wildcard.translate=_translateW
  Element.translate=_translateE
  All.translate=_translateA


# $Log: FSM.py,v $
# Revision 1.12  2004/05/20 09:50:20  ht
# remove some performance bottlenecks
#
# Revision 1.11  2004/05/18 18:02:07  ht
# changed min check to plain guard
#
# Revision 1.10  2004/05/17 17:31:15  ht
# plus-edge approach to numeric exponents working?
#
# Revision 1.9  2004/05/16 16:41:19  ht
# cul-de-sac wrt counters
#
# Revision 1.6  2004/04/01 13:31:43  ht
# work on final/block a bit
#
# Revision 1.5  2003/12/04 10:52:42  ht
# handle display of subsumption failures locally
#
# Revision 1.4  2002/11/25 11:39:35  ht
# protect against missing decl
#
# Revision 1.3  2002/11/11 18:18:40  ht
# use unicode properly for names/dumping
#
# Revision 1.2  2002/11/05 14:18:22  ht
# package FSM construction and checking and move to FSM
#
# Revision 1.1  2002/06/28 09:40:22  ht
# XSV as package: components
#
# Revision 1.2  2002/06/10 14:38:53  ht
# fix uba
#
# Revision 1.1  2002/05/24 22:33:14  ht
# split out of XMLSchema
#
